#
#   HorusLib - Decoder Delegate Functions
#

import struct
import time

from .payloads import HORUS_PAYLOAD_LIST, HORUS_CUSTOM_FIELDS

# Payload ID

def decode_payload_id(data: int) -> str:
    """
    Attempt to decode a payload ID into a callsign string.
    """

    if type(data) != int:
        return ValueError("payload_id - Invalid input type.")

    if data in HORUS_PAYLOAD_LIST:
        _str = HORUS_PAYLOAD_LIST[data]
    else:
        _str = "UNKNOWN_PAYLOAD_ID"

    return (_str, _str)


# Time representations

def decode_time_hms(data: bytes) -> str:
    """
    Decode a time field, encoded as three bytes representing hours, minutes and seconds of the current UTC day.

    Returns: String, as "HH:MM:SS"

    Example: \x01\x02\x03 -> "01:02:03"
    """

    if len(data) != 3:
        raise ValueError(f"time_hms - Input has incorrect length ({len(data)}), should be 3.")
    
    _hour = int(data[0])
    _minute = int(data[1])
    _second = int(data[2])

    _str = f"{_hour:02d}:{_minute:02d}:{_second:02d}"

    return (_str, _str)


def decode_time_biseconds(data:int) -> str:
    """
    Decode a time field, encoded as a uint16, representing seconds since the start of the UTC day,
    divided by 2 ('biseconds')

    Returns: String, as "HH:MM:SS"

    Examples: 
        0 -> 00:00:00
        1 -> 00:00:02

    """
    
    if type(data) != int:
        raise ValueError("time_biseconds - Invalid input type.")

    if (data < 0) or data > 43200:
        raise ValueError("time_biseconds - Input of of range (0-43200)")

    _str = time.strftime("%H:%M:%S", time.gmtime(data*2))

    return (_str, _str)

# Latitude/Longitude representations

def decode_degree_float(data:float) -> str:
    """ 
    Convert a degree (latitude/longitude) field, provided as a float,
    to a string representation, with 6 decimal places. 
    """
    if type(data) != float:
        raise ValueError("decimal_degrees - Invalid input type.")

    return (data, f"{data:.6f}")


def decode_degree_fixed3(data:bytes) -> str:
    """ 
    Convert a degree (latitude/longitude) field, provided as a 
    three-byte fixed-point representation, to a string.

    The input is interpreted as the 3 most-significant-bytes of a
    little-endian 4-byte signed integer. The LSB is set to 0x00.

    Once converted to an int, the value is then scaled to degrees by
    multilying by 1e-7.

    """

    if type(data) != bytes:
        raise ValueError("degree_fixed3 - Invalid input type.")

    if len(data) != 3:
        raise ValueError("degree_fixed3 - Invalid input length.")

    # Add input onto a null byte
    _temp = b'\x00' + data

    # Parse as a signed int.
    _value = struct.unpack('<i', _temp)[0]
    _value_degrees = _value * 1e-7

    return (_value_degrees, f"{_value_degrees:.6f}")


def decode_battery_5v_byte(data: int) -> str:
    """
    Decode a battery voltage, encoded as as a single byte, where
    0 = 0v, 255 = 5.0V, with linear steps in between.
    """

    if type(data) != int:
        raise ValueError("battery_5v_byte - Invalid input type.")

    _batt = 5.0*data/255.0

    return (_batt, f"{_batt:.2f}")


delegate_list = {
    'payload_id': decode_payload_id,
    'time_hms': decode_time_hms,
    'time_biseconds': decode_time_biseconds,
    'degree_float': decode_degree_float,
    'degree_fixed3': decode_degree_fixed3,
    'battery_5v_byte': decode_battery_5v_byte
}

def decode_field(field_type:str, data):
    """ Attempt to decode a field, supplied as bytes, using a specified delegate function """

    if field_type in delegate_list:
        return delegate_list[field_type](data)
    else:
        if (field_type == 'none') or (field_type == 'None') or (field_type == None):
            # Basic datatype, just convert to a string using Pythons internal conversions.
            if (type(data) == float) or (type(data) == int) or (type(data) == str):
                return (data, f"{data}")
            else:
                raise ValueError(f"Data has unknown type ({str(type(data))}) and could not be decoded.")
        else:
            raise ValueError(f"Invalid field type - {field_type}")


def decode_custom_fields(data:bytes, payload_id:str):
    """ Attempt to decode custom field data from the 9-byte custom section of a 32-byte payload """

    if payload_id not in HORUS_CUSTOM_FIELDS:
        raise ValueError(f"Custom Field Decoder - Unknown payload ID {payload_id}")

    _custom_field = HORUS_CUSTOM_FIELDS[payload_id]
    _struct = _custom_field['struct']
    _struct_len = struct.calcsize(_struct)
    _field_names = _custom_field['fields']

    if type(data) != bytes:
        raise ValueError("Custom Field Decoder - Invalid Input type.")

    if len(data) !=_struct_len:
        raise ValueError(f"Custom Field Decoder - Invalid Input Length ({len(data)}, should be {_struct_len}).")

    # Attempt to parse the data.
    _raw_fields = struct.unpack(_struct, data)

    if len(_field_names) != len(_raw_fields):
        raise ValueError(f"Custom Field Decoder - Packet format defines {len(_field_names)} fields, got {len(_raw_fields)} from struct.")

    _output_fields = []
    _output_dict = {}
    for _i in range(len(_raw_fields)):
        _field_name = _field_names[_i][0]
        _field_type = _field_names[_i][1]
        _field_data = _raw_fields[_i]

        # Decode field to string.
        (_decoded, _decoded_str) = decode_field(_field_type, _field_data)

        _output_dict[_field_name] = _decoded

        _output_fields.append(_decoded_str)

    _output_fields_str = ",".join(_output_fields)

    return (_output_dict, _output_fields_str)




if __name__ == "__main__":

    tests = [
        ['time_hms', b'\x01\x02\x03', "01:02:03"],
        ['time_hms', b'\x17\x3b\x3b', "23:59:59"],
        ['time_biseconds', 0, "00:00:00"],
        ['time_biseconds', 1, "00:00:02"],
        ['time_biseconds', 43199, "23:59:58"],
        ['time_biseconds', 43200, "00:00:00"],
        ['degree_float', 0.0, "0.000000"],
        ['degree_float', 0.001, "0.001000"],
        ['degree_float', -34.01, "-34.010000"],
        ['degree_float', -138.000001, "-138.000001"],
        ['degree_fixed3', b'\x00\x00\x00', "0.0"],
        ['battery_5v_byte', 0, "0.00"],
        ['battery_5v_byte', 128, "2.51"],
        ['battery_5v_byte', 255, "5.00"],
    ]

    for _test in tests:
        _field_type = _test[0]
        _input = _test[1]
        _output = _test[2]

        _decoded = decode_field(_field_type, _input)
        print(f"{_field_type} {str(_input)} -> {_decoded}")
        assert(_decoded == _output)
    
    print("All tests passed!")


#
#   HorusLib - Payload ID List
#
import json
import logging
import requests

# Global payload list
HORUS_PAYLOAD_LIST = {0:'4FSKTEST', 1:'HORUSBINARY', 65535:'HORUSTEST'}

# URL for payload list
# TODO: Move this into horusdemodlib repo
PAYLOAD_ID_LIST_URL = "https://raw.githubusercontent.com/projecthorus/horusdemodlib/master/payload_id_list.txt"

# Custom field data. 
HORUS_CUSTOM_FIELDS = {
    "HORUSTEST": {
        "struct": "<BbBfH",
        "fields": [
            ["cutdown_battery_voltage", "battery_5v_byte"],
            ["external_temperature", "none"],
            ["test_counter", "none"],
            ["test_float_field", "none"],
            ["test_int_field", "none"]
        ]
    },
    "HORUSTEST2": {
        "struct": "<BbBH4x",
        "fields": [
            ["cutdown_battery_voltage", "battery_5v_byte"],
            ["external_temperature", "none"],
            ["test_counter", "none"],
            ["test_int_field", "none"]
        ]
    }
}


# Custom Field JSON URL
HORUS_CUSTOM_FIELD_URL = "https://raw.githubusercontent.com/projecthorus/horusdemodlib/master/custom_field_list.json"

def read_payload_list(filename="payload_id_list.txt"):
    """ Read in the payload ID list, and return the parsed data as a dictionary """

    # Dummy payload list.
    payload_list = HORUS_PAYLOAD_LIST

    try:
        with open(filename,'r') as file:
            for line in file:
                # Skip comment lines.
                if line[0] == '#':
                    continue
                else:
                    # Attempt to split the line with a comma.
                    _params = line.split(',')
                    if len(_params) != 2:
                        # Invalid line.
                        logging.error("Could not parse line: %s" % line)
                    else:
                        try:
                            _id = int(_params[0])
                            _callsign = _params[1].strip()
                            payload_list[_id] = _callsign
                        except:
                            logging.error("Error parsing line: %s" % line)
    except Exception as e:
        logging.error("Error reading Payload ID list, does it exist? - %s" % str(e))

    logging.debug("Known Payload IDs:")
    for _payload in payload_list:
        logging.debug("\t%s - %s" % (_payload, payload_list[_payload]))

    return payload_list


def grab_latest_payload_id_list(url=PAYLOAD_ID_LIST_URL, local_file="payload_id_list.txt"):
    """ Attempt to download the latest payload ID list from Github """

    # Download the list.
    try:
        logging.info("Attempting to download latest payload ID list from GitHub...")
        _r = requests.get(url, timeout=10)
    except Exception as e:
        logging.error("Unable to get latest payload ID list: %s" % str(e))
        return False

    # Check it is what we think it is..
    if "HORUS BINARY PAYLOAD ID LIST" not in _r.text:
        logging.error("Downloaded payload ID list is invalid.")
        return False

    # So now we most likely have a valid payload ID list, so write it out.
    with open(local_file, 'w') as f:
        f.write(_r.text)

    logging.info("Updated payload ID list successfully!")
    return True


def init_payload_id_list():
    """ Initialise and update the local payload ID list. """

    grab_latest_payload_id_list()
    HORUS_PAYLOAD_LIST = read_payload_list()



def read_custom_field_list(filename="custom_field_list.json"):
    """ 
    Read in a JSON file containing descriptions of custom payload fields,
    for use with the Horus Binary v2 32-byte payload format.
    """

    _custom_field_list = HORUS_CUSTOM_FIELDS

    try:
        # Read in entirity of file contents.
        _f = open(filename, 'r')
        _raw_data = _f.read()
        _f.close()

        # Attempt to parse JSON
        _field_data = json.loads(_raw_data)

        if type(_field_data) != dict:
            logging.error("Error reading custom field list, using defaults.")
            return _custom_field_list
        
        # Iterate through fields in the file we just read in
        for _payload in _field_data:
            _data = _field_data[_payload]

            if ("struct" in _data) and ("fields" in _data):
                _custom_field_list[_payload] = {
                    "struct": _data["struct"],
                    "fields": _data["fields"]
                }
                logging.debug(f"Loaded custom field data for {_payload}.")
        
        return _custom_field_list

    except Exception as e:
        logging.error(f"Error parsing custom field list file ({filename}): {str(e)}")
        return _custom_field_list



def grab_latest_custom_field_list(url=HORUS_CUSTOM_FIELD_URL, local_file="custom_field_list.json"):
    """ Attempt to download the latest custom field list from Github """

    # Download the list.
    try:
        logging.info("Attempting to download latest custom field list from GitHub...")
        _r = requests.get(url, timeout=10)
    except Exception as e:
        logging.error("Unable to get latest custom field list: %s" % str(e))
        return False

    # Check it is what we think it is..
    # (Currently checking for the presence of one of the test payloads)
    if "HORUSTEST" not in _r.text:
        logging.error("Downloaded custom field list is invalid.")
        return False

    # So now we most likely have a valid custom field list, so write it out.
    with open(local_file, 'w') as f:
        f.write(_r.text)

    logging.info("Updated custom field list successfully!")
    return True


def init_custom_field_list():
    """ Initialise and update the local custom field list """
    grab_latest_custom_field_list()
    HORUS_CUSTOM_FIELDS = read_custom_field_list()


if __name__ == "__main__":

    # Setup Logging
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s", level=logging.DEBUG
    )

    init_payload_id_list()
    print(HORUS_PAYLOAD_LIST)

    init_custom_field_list()
    print(HORUS_CUSTOM_FIELDS)


#
#   HorusLib - Checksumming functions
#

import crcmod
import logging
import struct



def ukhas_crc(data:bytes) -> str:
    """
    Calculate the CRC16 CCITT checksum of *data*.
    
    (CRC16 CCITT: start 0xFFFF, poly 0x1021)
    """
    crc16 = crcmod.predefined.mkCrcFun('crc-ccitt-false')

    return hex(crc16(data))[2:].upper().zfill(4)


def check_packet_crc(data:bytes, checksum:str='crc16'):
    """ 
    Attempt to validate a packets checksum, which is assumed to be present
    in the last few bytes of the packet.

    Support CRC types: CRC16-CCITT

    """

    if (checksum == 'crc16') or (checksum == 'CRC16') or (checksum == 'crc16-ccitt') or (checksum == 'CRC16-CCITT'):
        # Check we have enough data for a sane CRC16.
        if len(data) < 3:
            raise ValueError(f"Checksum - Not enough data for CRC16!")

        # Decode the last 2 bytes as a uint16
        _packet_checksum = struct.unpack('<H', data[-2:])[0]

        # Calculate a CRC over the rest of the data
        _crc16 = crcmod.predefined.mkCrcFun('crc-ccitt-false')
        _calculated_crc = _crc16(data[:-2])

        if _calculated_crc == _packet_checksum:
            return True
        else:
            logging.debug(f"Calculated: {hex(_calculated_crc)}, Packet: {hex(_packet_checksum)}")
            return False

    else:
        raise ValueError(f"Checksum - Unknown Checksym type {checksum}.")



if __name__ == "__main__":

    # Setup Logging
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s", level=logging.DEBUG
    )

    tests = [
        ['crc16', b'\x01\x12\x00\x00\x00\x23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1C\x9A\x95\x45', True],
        ['crc16', b'\x01\x12\x00\x00\x00\x23\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x1C\x9A\x95\x45', False],
        ['crc16', b'\x01\x12\x02\x00\x02\xbc\xeb!AR\x10\x00\xff\x00\xe1\x7e', True],
        #           id      seq_no  HH   MM  SS  lat             lon             alt    spd  sat tmp bat custom data
        ['crc16', b'\xFF\xFF\x12\x00\x00\x00\x23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe8\x82', True],
    ]

    for _test in tests:
        _format = _test[0]
        _input = _test[1]
        _output = _test[2]

        _decoded = check_packet_crc(_input, _format)

        print(f"Packet: {_input}. CRC OK: {_decoded}")
        assert(_decoded == _output)

    print("All tests passed!")


#!/usr/bin/env python3
#
#   HorusDemodLib - LibHorus Wrapper Class
#

import audioop
import ctypes
from ctypes import *
import logging
import sys
from enum import Enum
import os
import logging
from .decoder import decode_packet, hex_to_bytes

MODEM_STATS_NR_MAX = 8
MODEM_STATS_NC_MAX = 50
MODEM_STATS_ET_MAX = 8
MODEM_STATS_EYE_IND_MAX = 160
MODEM_STATS_NSPEC = 512
MODEM_STATS_MAX_F_EST = 4


class COMP(Structure):
    """
    Used in MODEM_STATS for representing IQ.
    """
    _fields_ = [
        ("real", c_float),
        ("imag", c_float)
    ]


class MODEM_STATS(Structure):  # modem_stats.h
    """
    Extended modem stats structure
    """
    _fields_ = [
        ("Nc", c_int),
        ("snr_est", c_float),
        # rx_symbols[MODEM_STATS_NR_MAX][MODEM_STATS_NC_MAX+1];
        ("rx_symbols", (COMP * MODEM_STATS_NR_MAX)*(MODEM_STATS_NC_MAX+1)),
        ("nr", c_int),
        ("sync", c_int),
        ("foff", c_float),
        ("rx_timing", c_float),
        ("clock_offset", c_float),
        ("sync_metric", c_float),
        # float  rx_eye[MODEM_STATS_ET_MAX][MODEM_STATS_EYE_IND_MAX];
        ("rx_eye", (c_float * MODEM_STATS_ET_MAX)*MODEM_STATS_EYE_IND_MAX),
        ("neyetr", c_int),
        ("neyesamp", c_int),
        ("f_est", c_float*MODEM_STATS_MAX_F_EST),
        ("fft_buf", c_float * 2*MODEM_STATS_NSPEC),
        ("fft_cfg", POINTER(c_ubyte))
    ]


class Mode(Enum):
    """
    Modes (and aliases for modes) for the HorusLib modem
    """
    BINARY = 0
    BINARY_V1 = 0
    RTTY_7N2 = 90
    RTTY = 90
    BINARY_V2_256BIT = 1
    BINARY_V2_128BIT = 2


class Frame():
    """
    Frame class used for demodulation attempts. 

    Attributes
    ----------

    data : bytes
        Demodulated data output. Empty if demodulation didn't succeed
    sync : bool
        Modem sync status
    snr : float
        Estimated SNR
    crc_pass : bool
        CRC check status
    extended_stats
        Extended modem stats. These are provided as c_types so will need to be cast prior to use. See MODEM_STATS for structure details
    """

    def __init__(self, data: bytes, sync: bool, crc_pass: bool, snr: float, extended_stats: MODEM_STATS):
        self.data = data
        self.sync = sync
        self.snr = snr
        self.crc_pass = crc_pass
        self.extended_stats = extended_stats


class HorusLib():
    """
    HorusLib provides a binding to horuslib to demoulate frames.

    Example usage:

    from horuslib import HorusLib, Mode
    with HorusLib(, mode=Mode.BINARY, verbose=False) as horus:
        with open("test.wav", "rb") as f:
            while True:
                data = f.read(horus.nin*2)
                if horus.nin != 0 and data == b'': #detect end of file
                    break
                output = horus.demodulate(data)
                if output.crc_pass and output.data:
                    print(f'{output.data.hex()} SNR: {output.snr}')
                    for x in range(horus.mfsk):
                        print(f'F{str(x)}: {float(output.extended_stats.f_est[x])}')

    """

    def __init__(
        self,
        libpath=f"",
        mode=Mode.BINARY,
        rate=-1,
        tone_spacing=-1,
        stereo_iq=False,
        verbose=False,
        callback=None,
        sample_rate=48000
    ):
        """
        Parameters
        ----------
        libpath : str
            Path to libhorus
        mode : Mode
            horuslib.Mode.BINARY, horuslib.Mode.BINARY_V2_256BIT, horuslib.Mode.BINARY_V2_128BIT, horuslib.Mode.RTTY, RTTY_7N2 = 99
        rate : int
            Changes the modem rate for supported modems. -1 for default
        tone_spacing : int
            Spacing between tones (hz) -1 for default
        stereo_iq : bool
            use stereo (IQ) input (quadrature)
        verbose : bool
            Enabled horus_set_verbose
        callback : function
            When set you can use add_samples to add any number of audio frames and callback will be called when a demodulated frame is avaliable.
        sample_rate : int
            The input sample rate of the audio input
        """

        if sys.platform == "darwin":
            libpath = os.path.join(libpath, "libhorus.dylib")
        elif sys.platform == "win32":
            libpath = os.path.join(libpath, "libhorus.dll")
        else:
            libpath = os.path.join(libpath, "libhorus.so")

        # future improvement would be to try a few places / names
        self.c_lib = ctypes.cdll.LoadLibrary(libpath)

        # horus_open_advanced
        self.c_lib.horus_open_advanced.restype = POINTER(c_ubyte)

        # horus_nin
        self.c_lib.horus_nin.restype = c_uint32

        # horus_get_Fs
        self.c_lib.horus_get_Fs.restype = c_int

        # horus_set_freq_est_limits - (struct horus *hstates, float fsk_lower, float fsk_upper)
        self.c_lib.horus_set_freq_est_limits.argtype = [
            POINTER(c_ubyte),
            c_float,
            c_float,
        ]

        # horus_get_max_demod_in
        self.c_lib.horus_get_max_demod_in.restype = c_int

        # horus_get_max_ascii_out_len
        self.c_lib.horus_get_max_ascii_out_len.restype = c_int

        # horus_crc_ok
        self.c_lib.horus_crc_ok.restype = c_int

        # horus_get_modem_extended_stats - (struct horus *hstates, struct MODEM_STATS *stats)
        self.c_lib.horus_get_modem_extended_stats.argtype = [
            POINTER(MODEM_STATS),
            POINTER(c_ubyte),
        ]

        # horus_get_mFSK
        self.c_lib.horus_get_mFSK.restype = c_int

        # horus_rx
        self.c_lib.horus_rx.restype = c_int

        if type(mode) != type(Mode(0)):
            raise ValueError("Must be of type horuslib.Mode")
        else:
            self.mode = mode

        self.stereo_iq = stereo_iq

        self.callback = callback

        self.input_buffer = bytearray(b"")

        # intial nin
        self.nin = 0

        # try to open the modem and set the verbosity
        self.hstates = self.c_lib.horus_open_advanced(
            self.mode.value, rate, tone_spacing
        )
        self.c_lib.horus_set_verbose(self.hstates, int(verbose))

        # check that the modem was actually opened and we don't just have a null pointer
        if bool(self.hstates):
            logging.debug("Opened Horus API")
        else:
            logging.error("Couldn't open Horus API for some reason")
            raise EnvironmentError("Couldn't open Horus API")

        # build some class types to fit the data for demodulation using ctypes
        max_demod_in = int(self.c_lib.horus_get_max_demod_in(self.hstates))
        max_ascii_out = int(self.c_lib.horus_get_max_ascii_out_len(self.hstates))
        self.DemodIn = c_short * (max_demod_in * (1 + int(self.stereo_iq)))
        self.DataOut = c_char * max_ascii_out
        self.c_lib.horus_rx.argtype = [
            POINTER(c_ubyte),
            c_char * max_ascii_out,
            c_short * max_demod_in,
            c_int,
        ]

        self.mfsk = int(self.c_lib.horus_get_mFSK(self.hstates))

        self.resampler_state = None
        self.audio_sample_rate = sample_rate
        self.modem_sample_rate = 48000

    # in case someone wanted to use `with` style. I'm not sure if closing the modem does a lot.
    def __enter__(self):
        return self

    def __exit__(self, *a):
        self.close()

    def close(self) -> None:
        """
        Closes Horus modem.
        """
        self.c_lib.horus_close(self.hstates)
        logging.debug("Shutdown horus modem")

    def _update_nin(self) -> None:
        """
        Updates nin. Called every time RF is demodulated and doesn't need to be run manually
        """
        new_nin = int(self.c_lib.horus_nin(self.hstates))
        if self.nin != new_nin:
            logging.debug(f"Updated nin {new_nin}")
        self.nin = new_nin

    def demodulate(self, demod_in: bytes) -> Frame:
        """
        Demodulates audio in, into bytes output.

        Parameters
        ----------
        demod_in : bytes
            16bit, signed for audio in. You'll need .nin frames in to work correctly.
        """
        # resample to 48khz
        (demod_in, self.resampler_state) = audioop.ratecv(demod_in, 2, 1+int(self.stereo_iq), self.audio_sample_rate, self.modem_sample_rate, self.resampler_state)

        # from_buffer_copy requires exact size so we pad it out.
        buffer = bytearray(
            len(self.DemodIn()) * sizeof(c_short)
        )  # create empty byte array
        buffer[: len(demod_in)] = demod_in  # copy across what we have

        modulation = self.DemodIn  # get an empty modulation array
        modulation = modulation.from_buffer_copy(
            buffer
        )  # copy buffer across and get a pointer to it.

        data_out = self.DataOut()  # initilize a pointer to where bytes will be outputed

        self.c_lib.horus_rx(self.hstates, data_out, modulation, int(self.stereo_iq))

        stats = MODEM_STATS()
        self.c_lib.horus_get_modem_extended_stats(self.hstates, byref(stats))

        crc = bool(self.c_lib.horus_crc_ok(self.hstates))

        data_out = bytes(data_out)
        self._update_nin()

        # strip the null terminator out
        data_out = data_out[:-1]

        if data_out == bytes(len(data_out)):
            data_out = (
                b""  # check if bytes is just null and return an empty bytes instead
            )
        elif self.mode != Mode.RTTY_7N2:
            try:
                data_out = bytes.fromhex(data_out.decode("ascii"))
            except ValueError:
                logging.debug(data_out)
                logging.error("Couldn't decode the hex from the modem")
                return bytes()
        else:
            # Ascii
            data_out = data_out.decode("ascii")
            # Strip of all null characters.
            data_out = data_out.rstrip('\x00')

        frame = Frame(
            data=data_out,
            snr=float(stats.snr_est),
            sync=bool(stats.sync),
            crc_pass=crc,
            extended_stats=stats,
        )
        return frame
    
    def add_samples(self, samples: bytes):
        """ Add samples to a input buffer, to pass on to demodulate when we have nin samples """

        # Add samples to input buffer
        self.input_buffer.extend(samples)

        _processing = True
        _frame = None
        while _processing:
            # Process data until we have less than _nin samples.
            _nin = int(self.nin*(self.audio_sample_rate/self.modem_sample_rate))
            if len(self.input_buffer) > (_nin * 2):
                # Demodulate
                _frame = self.demodulate(self.input_buffer[:(_nin*2)])

                # Advance sample buffer.
                self.input_buffer = self.input_buffer[(_nin*2):]

                # If we have decoded a packet, send it on to the callback
                if len(_frame.data) > 0:
                    if self.callback:
                        self.callback(_frame)
            else:
                _processing = False
        
        return _frame


if __name__ == "__main__":
    import sys
    if len(sys.argv) != 2:
        raise ArgumentError("Usage python3 -m horusdemodlib.demod filename")
    filename = sys.argv[1]

    def frame_callback(frame):
        print(f"Callback: {frame.data.hex()} SNR: {frame.snr}")

        try:
            _packet = decode_packet(frame.data)
            print(f"Decoded Packet: {_packet['ukhas_str']}")
        except Exception as e:
            print(f"Error decoding packet: {str(e)}.")


    # Setup Logging
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s", level=logging.DEBUG
    )

    with HorusLib(mode=Mode.BINARY, verbose=False, callback=frame_callback, sample_rate=8000) as horus:
        with open(filename, "rb") as f:
            while True:
                # Fixed read size - 2000 samples
                data = f.read(2000 * 2)
                if horus.nin != 0 and data == b"":  # detect end of file
                    break
                output = horus.add_samples(data)
                if output:
                    print(f"Sync: {output.sync}  SNR: {output.snr}")



#
#   HorusLib - Binary Packet Decoder Functions
#
import codecs
import struct
import time
from .delegates import *
from .checksums import *
from .payloads import HORUS_CUSTOM_FIELDS, HORUS_PAYLOAD_LIST, init_custom_field_list, init_payload_id_list



#
#   Horus Binary V1 and V2 Packet Formats
#
HORUS_PACKET_FORMATS = {
    'horus_binary_v1': {
        'name': 'Horus Binary v1 22 Byte Format',
        'length': 22,
        'struct': '<BH3sffHBBbBH',
        'checksum': 'crc16',
        'fields': [
            ['payload_id', 'payload_id'],
            ['sequence_number', 'none'],
            ['time', 'time_hms'],
            ['latitude', 'degree_float'],
            ['longitude', 'degree_float'],
            ['altitude', 'none'],
            ['speed', 'none'],
            ['satellites', 'none'],
            ['temperature', 'none'],
            ['battery_voltage', 'battery_5v_byte'],
            ['checksum', 'none']
        ]
    },
    'horus_binary_v2_16byte': {
        'name': 'Horus Binary v2 16 Byte Format',
        'length': 16,
        'struct': '<BBH3s3sHBBH',
        'checksum': 'crc16',
        'fields': [
            ['payload_id', 'payload_id'],
            ['sequence_number', 'none'],
            ['time', 'time_biseconds'],
            ['latitude', 'degree_fixed3'],
            ['longitude', 'degree_fixed3'],
            ['altitude', 'none'],
            ['battery_voltage', 'battery_5v_byte'],
            ['flags', 'none'],
            ['checksum', 'none']
        ]
    },
    'horus_binary_v2_32byte': {
        'name': 'Horus Binary v2 32 Byte Format',
        'length': 32,
        'struct': '<HH3sffHBBbB9sH',
        'checksum': 'crc16',
        'fields': [
            ['payload_id', 'payload_id'],
            ['sequence_number', 'none'],
            ['time', 'time_hms'],
            ['latitude', 'degree_float'],
            ['longitude', 'degree_float'],
            ['altitude', 'none'],
            ['speed', 'none'],
            ['satellites', 'none'],
            ['temperature', 'none'],
            ['battery_voltage', 'battery_5v_byte'],
            ['custom', 'custom'],
            ['checksum', 'none']
        ]
    }
}

# Lookup for packet length to the appropriate format.
HORUS_LENGTH_TO_FORMAT = {
    22: 'horus_binary_v1',
    16: 'horus_binary_v2_16byte',
    32: 'horus_binary_v2_32byte'
}

def decode_packet(data:bytes, packet_format:dict = None) -> dict:
    """ 
    Attempt to decode a set of bytes based on a provided packet format.

    """

    if packet_format is None:
        # Attempt to lookup the format based on the length of the data if it has not been provided.
        if len(data) in HORUS_LENGTH_TO_FORMAT:
            packet_format = HORUS_PACKET_FORMATS[HORUS_LENGTH_TO_FORMAT[len(data)]]


    # Output dictionary
    _output = {
        'packet_format': packet_format,
        'crc_ok': False,
        'payload_id': 0
        }
    
    # Check the length provided in the packet format matches up with the length defined by the struct.
    _struct_length = struct.calcsize(packet_format['struct'])
    if _struct_length != packet_format['length']:
        raise ValueError(f"Decoder - Provided length {packet_format['length']} and struct length ({_struct_length}) do not match!")
    
    # Check the length of the input data bytes matches that of the struct.
    if len(data) != _struct_length:
        raise ValueError(f"Decoder - Input data has length {len(data)}, should be length {_struct_length}.")

    # Check the Checksum
    _crc_ok = check_packet_crc(data, checksum=packet_format['checksum'])

    if not _crc_ok:
        raise ValueError("Decoder - CRC Failure.")
    else:
        _output['crc_ok'] = True

    # Now try and decode the data.
    _raw_fields = struct.unpack(packet_format['struct'], data)

    # Check the number of decoded fields is equal to the number of field definitions in the packet format.
    if len(_raw_fields) != len(packet_format['fields']):
        raise ValueError(f"Decoder - Packet format defines {len(packet_format['fields'])} fields, got {len(_raw_fields)} from struct.")

    # Now we can start extracting and formatting fields.
    
    _ukhas_fields = []
    for _i in range(len(_raw_fields)):
        _field_name = packet_format['fields'][_i][0]
        _field_type = packet_format['fields'][_i][1]
        _field_data = _raw_fields[_i]


        if _field_name == 'custom':
            # Attempt to interpret custom fields.
            # Note: This requires that the payload ID has been decoded prior to this field being parsed.
            if _output['payload_id'] in HORUS_CUSTOM_FIELDS:
                (_custom_data, _custom_str) = decode_custom_fields(_field_data, _output['payload_id'])

                # Add custom fields to string
                _ukhas_fields.append(_custom_str)

                # Add custom fields to output dict.
                for _field in _custom_data:
                    _output[_field] = _custom_data[_field]

        # Ignore checksum field. (and maybe other fields?)
        elif _field_name not in ['checksum']:
            # Decode field to string.
            (_decoded, _decoded_str) = decode_field(_field_type, _field_data)

            _output[_field_name] = _decoded

            _ukhas_fields.append(_decoded_str)


    # Convert to a UKHAS-compliant string.
    _ukhas_str = ",".join(_ukhas_fields)
    _ukhas_crc = ukhas_crc(_ukhas_str.encode('ascii'))
    _output['ukhas_str'] = "$$" + _ukhas_str + "*" + _ukhas_crc

    return _output


def hex_to_bytes(data:str) -> bytes:
    """ Convert a string of hexadeximal digits to a bytes representation """
    try:
        _binary_string = codecs.decode(data, 'hex')
        return _binary_string
    except TypeError as e:
        logging.error("Error parsing line as hexadecimal (%s): %s" % (str(e), data))
        return None


if __name__ == "__main__":
    import argparse
    import sys


    # Read command-line arguments
    parser = argparse.ArgumentParser(description="Project Horus Binary Telemetry Decoder", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--test", action="store_true", default=False, help="Run unit tests.")
    parser.add_argument("--update", action="store_true", default=False, help="Download latest payload ID and custom fields files before continuing.")
    parser.add_argument("--decode", type=str, default=None, help="Attempt to decode a hexadecial packet.")
    parser.add_argument("-v", "--verbose", action="store_true", default=False, help="Verbose output (set logging level to DEBUG)")
    args = parser.parse_args()

    if args.verbose:
        _log_level = logging.DEBUG
    else:
        _log_level = logging.INFO

    # Setup Logging
    logging.basicConfig(
        format="%(asctime)s %(levelname)s: %(message)s", level=_log_level
    )

    if args.update:
        init_payload_id_list()
        init_custom_field_list()
    
    if args.decode is not None:
        try:
            _decoded = decode_packet(hex_to_bytes(args.decode))
            print(f"Decoded UKHAS String: {_decoded['ukhas_str']}")
        except ValueError as e:
            print(f"Error while decoding: {str(e)}")


    if args.test:

        tests = [
            ['horus_binary_v1', b'\x01\x12\x00\x00\x00\x23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x1C\x9A\x95\x45', ''],
            ['horus_binary_v1', b'\x01\x12\x00\x00\x00\x23\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x1C\x9A\x95\x45', 'error'],
            ['horus_binary_v2_16byte', b'\x01\x12\x02\x00\x02\xbc\xeb!AR\x10\x00\xff\x00\xe1\x7e', ''],
            #                             id      seq_no  HH   MM  SS  lat             lon            alt     spd sat tmp bat custom data -----------------------| crc16
            ['horus_binary_v2_32byte', b'\xFF\xFF\x12\x00\x00\x00\x23\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xe8\x82', '']
        ]

        for _test in tests:
            _format = _test[0]
            _input = _test[1]
            _output = _test[2]

            try:
                _decoded = decode_packet(_input)
                print(f"Input ({_format}): {str(_input)} - Output: {_decoded['ukhas_str']}")
                print(_decoded)
                # Insert assert checks here.

            except ValueError as e:
                print(f"Input ({_format}): {str(_input)} - Caught Error: {str(e)}")
                assert(_output == 'error')


        print("All tests passed!")

